from math import ceil
from functools import partial
from concurrent.futures import BrokenExecutor
from typing import Generic, TypeVar, TYPE_CHECKING, get_args, cast
from collections.abc import Sized, Iterable, Iterator, Collection, Callable

from .bfr import BufferedFutureResolver, KeyedFuture
from .args import Args
from .utils import class_logger, batches
from .typing import ReturnKey, Error
from ._registry import RegistryKey
from ._workerapi import WorkerAPI, ErrorResult
if TYPE_CHECKING:
    from .workforce import Workforce

T = TypeVar("T")


class Mapping(Generic[T]):
    """
    Represents the results of mapping a single function over some inputs.

    This class is given a Workforce and performs the actions needed to map a function over the inputs.
    This includes registering the function for use by the workers, batching the inputs, submitting the tasks
    and using a BufferedFutureResolver to resolve the submitted Futures.

    The results are accessed by iterating over the Mapping.

    Parameters
    ----------
    workforce : workforce.Workforce
        The Workforce in charge of executing the function.
    func : callable
        The function to map.
    inputs : iterable
        An iterable yielding arguments to `func`. Each item can be:
            - An args.Args object, having args and kwargs members.
            - A tuple, which will be unpacked when invoking `func` and used as positional arguments.
            - Any other object, which will be used as the single argument to `func`.
        If the iterable is Sized (i.e. has a length), then its length is also the length of this Mapping.
    ordered : bool, default False
        Whether the order of the outputs should correlate to the order of the `inputs`.
    buffering : int, optional
        Limits the number of `inputs` extracted before waiting for tasks to be completed.
        By default, no limit is set.
    batch_size : int, default 1
        Divide the `inputs` into batches that are sent to the workers instead of sending each input separately.
        This can boost running time significantly, especially when `func` is quick and there are many `inputs`.
    errors : {'raise', 'return', 'ignore', 'log'}, default 'raise'
        How to handle exceptions raised by `func`.
        - 'raise':  Raise the exception.
        - 'return': Return the exception object as if it were one of the results.
        - 'ignore': Ignore the exception completely. If an exception is encountered, the number of results
                    will be smaller than the number of `inputs`.
        - 'log':    Similar to 'ignore', but also log the exception.
    timeout : float, optional
        A global timeout, in seconds, on the entire calculation.
        The clock starts as soon as iteration over the results begins.
    return_key : {'none', 'idx', 'input'}, default 'none'
        Return with each result a key identifying it as a tuple (key, result). May be useful if `ordered` is False.
        - 'none':  Do not return a key. Each result is yielded as is.
        - 'idx':   The key is the zero-based running index of the input.
        - 'input': The key is the input itself in its entirety.
    """
    logger = class_logger()

    def __init__(self, workforce: "Workforce", func: Callable, inputs: Iterable,
                 ordered: bool = False, buffering: int | None = None, batch_size: int = 1, errors: Error = 'raise',
                 timeout: float | None = None, return_key: ReturnKey = 'none'):
        assert errors in get_args(Error)
        assert return_key in get_args(ReturnKey)
        self.batch_size = batch_size
        self.errors = errors
        self.return_key = return_key
        func_key = workforce.func_registry.register(func)
        self._n_inputs = len(inputs) if isinstance(inputs, Sized) else None
        if self._n_inputs and workforce.n_workers != 0:
            self.batch_size = min(self.batch_size, ceil(self._n_inputs / workforce.n_workers))
        assert self.batch_size >= 1
        if buffering is not None:
            buffering = ceil(buffering / self.batch_size)
        input_batches = self._batch_inputs(map(Args.canonize, inputs))
        self.bfr = BufferedFutureResolver(
            keyed_futures=self._generate_keyed_futures(workforce, func_key, input_batches),
            abort_on_error=True,
            ordered=ordered,
            buffering=buffering,
            timeout=timeout,
            name=f"Workforce-{workforce.workforce_id}-BFR",
        )
        results = self.bfr
        results = self._unbatch_results(results)
        results = self._handle_errors(results)
        results = self._handle_workforce_failures(results, workforce)
        if self.return_key == 'none':
            results = (i[1] for i in results)
        self._results = cast(Iterator[T], results)

    def __len__(self) -> int:
        if self._n_inputs is None:
            raise TypeError("input iterable has no length")
        return self._n_inputs

    def __iter__(self) -> Iterator[T]:
        return self._results

    def _batch_inputs(self, inputs: Iterable[Args]) -> Iterable[list[Args]]:
        return map(list, batches(inputs, self.batch_size))

    def _generate_keyed_futures(self, workforce: "Workforce", func_key: RegistryKey,
                                input_batches: Iterable[list[Args]]) -> Iterable[KeyedFuture[list]]:
        if not workforce.active:
            return
        make_keys = partial(self._make_keys, return_key=self.return_key, batch_size=self.batch_size)
        # Don't keep a reference so that the Mapping could be deleted before the generator is fully consumed
        del self
        for batch_idx, input_batch in enumerate(input_batches):
            future = workforce.submit(WorkerAPI.execute, workforce.workforce_id, func_key, input_batch)
            if future is not None:
                yield KeyedFuture(make_keys(batch_idx, input_batch), future)
            if not workforce.active:
                return

    # This must be static so that _generate_keyed_futures doesn't keep a reference to self
    @staticmethod
    def _make_keys(batch_idx: int, input_batch: Collection[Args], return_key: ReturnKey, batch_size: int) -> list:
        if return_key == 'idx':
            return list(range(s := batch_idx * batch_size, s + len(input_batch)))
        elif return_key == 'input':
            return [args.raw for args in input_batch]
        else:
            return [None] * len(input_batch)

    @staticmethod
    def _unbatch_results(results):
        for batch_keys, batch_results in results:
            yield from zip(batch_keys, batch_results)

    def _handle_errors(self, results):
        for key, result in results:
            if isinstance(result, ErrorResult):
                exc = result.exc
                if self.errors == 'return':
                    yield key, exc
                elif self.errors == 'ignore':
                    continue
                elif self.errors == 'log':
                    self.logger.info(f'exception raised from key {key}', exc_info=exc)
                    continue
                else:
                    assert self.errors == 'raise'
                    self.terminate()
                    raise exc
            else:
                yield key, result

    def _handle_workforce_failures(self, results, workforce):
        try:
            yield from results
        except BrokenExecutor:
            self.logger.warning("Executor got broken while reading results, shutting down workforce and re-raising")
            workforce.shutdown()
            raise

    def terminate(self) -> None:
        self.bfr.terminate()


class SingularMapping(Mapping[T]):
    """
    A Mapping that is also in charge of shutting its Workforce down when finished.

    This Mapping is used when it is the only Mapping of a Workforce, which cannot be used as
    a context manager to allow proper shutdown, e.g. by the ultimap sugaring function.
    """
    def __init__(self, workforce: "Workforce", *args, **kwargs):
        super().__init__(workforce, *args, **kwargs)
        self.workforce = workforce

    def __iter__(self) -> Iterator[T]:
        try:
            yield from super().__iter__()
        finally:
            self.terminate()

    def terminate(self) -> None:
        self.workforce.shutdown()
        super().terminate()



import sys
import weakref
import threading
import concurrent.futures
from collections.abc import Callable, Iterable
from typing import Literal, TypeVar, get_args, overload, TypeAlias

from .args import Args
from .utils import SyncCounter, class_logger
from .typing import Error, ReturnKey
from .backend import get_backend, InlineBackend, BackendArgument
from ._workerapi import WorkerAPI
from ._registry import SerializedItemsRegistry
from ._mapping import Mapping, SingularMapping
from ._recursive import make_recursive


T = TypeVar("T")
S = TypeVar("S")
ShutdownMode: TypeAlias = Literal['auto', 'wait', 'nowait']
_ErrorNotReturn: TypeAlias = Literal['raise', 'ignore', 'log']


class Workforce:
    """
    A pool of workers capable of performing tasks.

    A Workforce is a collection of workers, running using a specific backend. They are spun up and initialized when
    the Workforce is initialized, and are then waiting for tasks to perform.
    Wraps a concurrent.futures.Executor object.

    Parameters
    ----------
    backend : str, Backend or Backend class, default 'multiprocessing'
        The backend to use for the workers. Can be either the name of a backend as a string, an instance of
        a Backend subclass or the Backend subclass itself. See backend.get_backend for details.
        Supported backends include 'multiprocessing', 'threading', 'inline', 'multiprocessing-spawn',
        'multiprocessing-fork' and 'multiprocessing-forkserver'.
    n_workers : int or float, optional
        The number of worker in the Workforce. Default and meaning of different argument types depend on `backend`:
        * multiprocessing:
          - Positive int:   Number of processes.
          - Negative int:   Refers to the number of CPUs (-1 means n_CPUs, -2 means 1 fewer, etc.)
          - Positive float: Fraction of the number of CPUs.
          - Default:        Total number of CPUs.
        * threading:
          - Positive int: Number of threads.
          - No default is provided; `n_workers` must be specified.
        * inline:
          - If provided, must be 0.
        If `n_workers` is zero, then `backend` is replaced with 'inline'.
    shutdown_mode : {'auto', 'wait', 'nowait'}, default 'auto'
        Upon shutdown, whether to wait for all running tasks to finish running. 'auto' means 'nowait' if possible
        (determined by the backend and version), otherwise 'wait'.

    Attributes
    ----------
    active : bool
        Whether the Workforce is alive and ready.
    executor : concurrent.futures.Executor
        Underlying Executor managing the workers.
    workforce_id : int
        Unique ID for this Workforce.
    func_registry : SerializedItemsRegistry
        A centralized location for functions to be registered by the Workforce and accessed by the workers.
    """
    logger = class_logger()
    workforce_id_counter = SyncCounter()

    def __init__(self, backend: BackendArgument = "multiprocessing", n_workers: int | float | None = None,
                 shutdown_mode: ShutdownMode = 'auto'):
        assert shutdown_mode in get_args(ShutdownMode)
        self.active = False
        self.executor = None
        self.workforce_id = self.workforce_id_counter.get_next_value()
        self._executor_lock = threading.Lock()
        self.backend = get_backend(backend)
        if n_workers == 0 and not isinstance(self.backend, InlineBackend):
            self.logger.warning(f"replacing backend {self.backend} with InlineBackend()")
            self.backend = InlineBackend()
        self.n_workers = self.backend.parse_n_workers(n_workers)
        self.shutdown_mode = shutdown_mode
        self.func_registry = SerializedItemsRegistry[Callable](self.backend)
        self._mappings: list[weakref.ReferenceType] = []
        self.executor = self.backend.Executor(
            max_workers=self.n_workers,
            initializer=WorkerAPI.initializer,
            initargs=WorkerAPI.init_args(self),
        )
        self.active = True

    @overload
    def map(self, func: Callable[..., T], inputs: Iterable, *, ordered: bool = ..., buffering: int | None = ...,
            batch_size: int = ..., errors: Literal['return'], timeout: float | None = ...,
            return_key: Literal['none'] = ..., recursive: bool = ...) -> Mapping[T | Exception]: ...

    @overload
    def map(self, func: Callable[..., T], inputs: Iterable, *, ordered: bool = ..., buffering: int | None = ...,
            batch_size: int = ..., errors: Literal['return'], timeout: float | None = ...,
            return_key: Literal['idx'], recursive: bool = ...) -> Mapping[tuple[int, T | Exception]]: ...

    @overload
    def map(self, func: Callable[..., T], inputs: Iterable[S], *, ordered: bool = ..., buffering: int | None = ...,
            batch_size: int = ..., errors: Literal['return'], timeout: float | None = ...,
            return_key: Literal['input'], recursive: bool = ...) -> Mapping[tuple[S, T | Exception]]: ...

    @overload
    def map(self, func: Callable[..., T], inputs: Iterable, *, ordered: bool = ..., buffering: int | None = ...,
            batch_size: int = ..., errors: _ErrorNotReturn = ..., timeout: float | None = ...,
            return_key: Literal['none'] = ..., recursive: bool = ...) -> Mapping[T]: ...

    @overload
    def map(self, func: Callable[..., T], inputs: Iterable, *, ordered: bool = ..., buffering: int | None = ...,
            batch_size: int = ..., errors: _ErrorNotReturn = ..., timeout: float | None = ...,
            return_key: Literal['idx'], recursive: bool = ...) -> Mapping[tuple[int, T]]: ...

    @overload
    def map(self, func: Callable[..., T], inputs: Iterable[S], *, ordered: bool = ..., buffering: int | None = ...,
            batch_size: int = ..., errors: _ErrorNotReturn = ..., timeout: float | None = ...,
            return_key: Literal['input'], recursive: bool = ...) -> Mapping[tuple[S, T]]: ...

    def map(self, func: Callable, inputs: Iterable, *, ordered: bool = False, buffering: int | None = None,
            batch_size: int = 1, errors: Error = 'raise', timeout: float | None = None,
            return_key: ReturnKey = 'none', recursive: bool = False) -> Mapping:
        """
        Map a function over several inputs, performing the tasks by the workers.

        The inputs can be buffered (i.e., not consumed all at once) and gathered into batches. There are several
        options for handling errors and a timeout for the entire calculation can be set.

        The work begins (and the timeout clock starts) only when iteration over the results begins. Results
        become available as they are ready.

        Parameters
        ----------
        func : callable
            The function to map over all `inputs`.
        inputs : iterable
            An iterable yielding arguments to `func`. Each item can be:
                - An args.Args object, having args and kwargs members.
                - A tuple, which will be unpacked when invoking `func` and used as positional arguments.
                - Any other object, which will be used as the single argument to `func`.
            If the iterable is Sized (i.e. has a length), then the returned Mapping is also Sized and its length
            is the same as that of `inputs`.
        ordered : bool, default False
            Whether the order of the outputs should correlate to the order of the `inputs`.
            Note: with `recursive=True`, the order is only guaranteed for the original inputs of the mapping.
        buffering : int, optional
            Limits the number of `inputs` extracted before waiting for tasks to be completed.
            By default, no limit is set.
        batch_size : int, default 1
            Divide the `inputs` into batches that are sent to the workers instead of sending each input separately.
            This can boost running time significantly, especially when `func` is quick and there are many `inputs`.
        errors : {'raise', 'return', 'ignore', 'log'}, default 'raise'
            How to handle exceptions raised by `func`.
            - 'raise':  Raise the exception.
            - 'return': Return the exception object as if it were one of the results.
            - 'ignore': Ignore the exception completely. If an exception is encountered, the number of results
                        will be smaller than the number of `inputs`.
            - 'log':    Similar to 'ignore', but also log the exception.
        timeout : float, optional
            A global timeout, in seconds, on the entire calculation.
            The clock starts as soon as iteration over the results begins.
        return_key : {'none', 'idx', 'input'}, default 'none'
            Return with each result a key identifying it as a tuple (key, result). May be useful if `ordered` is False.
            - 'none':  Do not return a key. Each result is yielded as is.
            - 'idx':   The key is the zero-based running index of the input.
            - 'input': The key is the input itself in its entirety.
        recursive : bool, default False
            Whether the mapping function may provide additional inputs for mapping.
            When True, the following changes take place:
            - An additional `add_input` callable argument is passed to the function. Call it for every additional input
              to be processed, providing either a single tuple (args), an `Args` instance, or *args and **kwargs.
            - When added inputs are available, they will generally be processed before any additional original inputs.
            - When `ordered` is True, only the order of the original inputs is guaranteed.
            - When using the inline backend, exceptions raised from the function may cause the mapping to hang.

        Returns
        -------
        Mapping
            Iterable object which yields the results of the operation upon iteration.
            If `input` is Sized, it is also Sized.
        """
        self._check_active()
        if recursive:
            func, inputs = make_recursive(func, inputs, self.backend)
        mapping = Mapping(self, func, inputs, ordered, buffering, batch_size, errors, timeout, return_key)
        self._mappings.append(weakref.ref(mapping))
        return mapping

    def do_one(self, func: Callable[..., T], args, *, timeout: float | None = None) -> T:
        """
        Perform a single task by a worker and return the result.

        Parameters
        ----------
        func : callable
            The function to be invoked by the worker.
        args : args.Args, tuple or any other object
            Arguments for `func`. Can be either of the following:
            - An args.Args object, having args and kwargs members.
            - A tuple, which will be unpacked when invoking `func` and used as positional arguments.
            - Any other object, which will be used as the single argument to `func`.
        timeout : float, optional
            The number of seconds to wait for the result. By default, there is no limit on the wait time.

        Returns
        -------
        The result of invoking `func` on the `args`.

        Raises
        ------
        TimeoutError
            If `timeout` was provided and the calculation timed out.
        RuntimeError
            If the Workforce has been shut down while submitting the task.
        """
        self._check_active()
        func_key = self.func_registry.register(func)
        future = self.submit(WorkerAPI.execute_one, self.workforce_id, func_key, Args.canonize(args))
        if future is None:
            raise RuntimeError("Workforce has been shut down during do_one execution")
        try:
            return future.result(timeout=timeout)
        except concurrent.futures.TimeoutError as exc:
            try:
                future.cancel()
            except:
                self.logger.exception("do_one future timed out but its cancellation raised an exception (suppressing)")
            raise TimeoutError("future timed out") from exc

    def __enter__(self) -> "Workforce":
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.logger.debug("Workforce exit handler called")
        self.shutdown()

    def __del__(self):
        self.shutdown()

    def shutdown(self, wait: bool | None = None) -> None:
        """
        Clean-up the resources associated with the Workforce.

        It is safe to call this method several times. However, no other methods can be called after this one.

        Parameters
        ----------
        wait : bool, optional
            If True then shutdown will not return until all running futures have finished executing.
            Default falls back to the `shutdown_mode` parameter of the Workforce.

        Returns
        -------
        None
        """
        self.active = False
        if self.executor is not None:
            with self._executor_lock:
                if self.executor is not None:
                    for mapping_ref in self._mappings:
                        if (mapping := mapping_ref()) is not None:
                            mapping.terminate()
                    wait = (
                        wait if wait is not None else
                        True if self.shutdown_mode == 'wait' else
                        False if self.shutdown_mode == 'nowait' else
                        not self.backend.executor_shutdown_nowait_allowed
                    )
                    try:
                        kwargs: dict[str, bool] = {}
                        if sys.version_info >= (3, 9):
                            kwargs['cancel_futures'] = True
                        self.executor.shutdown(wait=wait, **kwargs)
                    finally:
                        self.executor = None
                    try:
                        self.backend.shutdown()
                    finally:
                        self.backend = None

    def submit(self, func: Callable, *args, **kwargs) -> concurrent.futures.Future:
        """
        Submit a single task to be performed by a worker.

        This is a thin wrap around concurrent.futures.Executor.submit, providing a low-level interface.
        To submit a single task and simply get its result back, use `do_one`.
        To submit several tasks that all run the same function, use `map`.

        Parameters
        ----------
        func : callable
            The function to be invoked by the worker.
        *args
            Positional arguments for `func`.
        **kwargs
            Keyword arguments for `func`.

        Returns
        -------
        concurrent.futures.Future
            Object representing the result of the calculation.
        """
        try:
            with self._executor_lock:
                if self.active:
                    return self.executor.submit(func, *args, **kwargs)
                else:
                    raise RuntimeError("Submit called on an inactive workforce")
        except concurrent.futures.BrokenExecutor:
            self.logger.warning("Executor got broken during submission, shutting down and re-raising")
            self.shutdown()
            raise

    def _check_active(self) -> None:
        if not self.active:
            raise RuntimeError("This workforce instance has already been shut down")


@overload
def ultimap(func: Callable[..., T], inputs: Iterable, *, ordered: bool = ..., buffering: int | None = ...,
            batch_size: int = ..., errors: Literal['return'], timeout: float | None = ...,
            return_key: Literal['none'] = ..., recursive: bool = ..., backend: BackendArgument = ...,
            n_workers: int | float | None = ...,
            shutdown_mode: ShutdownMode = ...) -> SingularMapping[T | Exception]: ...


@overload
def ultimap(func: Callable[..., T], inputs: Iterable, *, ordered: bool = ..., buffering: int | None = ...,
            batch_size: int = ..., errors: Literal['return'], timeout: float | None = ...,
            return_key: Literal['idx'], recursive: bool = ..., backend: BackendArgument = ...,
            n_workers: int | float | None = ...,
            shutdown_mode: ShutdownMode = ...) -> SingularMapping[tuple[int, T | Exception]]: ...


@overload
def ultimap(func: Callable[..., T], inputs: Iterable[S], *, ordered: bool = ..., buffering: int | None = ...,
            batch_size: int = ..., errors: Literal['return'], timeout: float | None = ...,
            return_key: Literal['input'], recursive: bool = ..., backend: BackendArgument = ...,
            n_workers: int | float | None = ...,
            shutdown_mode: ShutdownMode = ...) -> SingularMapping[tuple[S, T | Exception]]: ...


@overload
def ultimap(func: Callable[..., T], inputs: Iterable, *, ordered: bool = ..., buffering: int | None = ...,
            batch_size: int = ..., errors: _ErrorNotReturn = ..., timeout: float | None = ...,
            return_key: Literal['none'] = ..., recursive: bool = ..., backend: BackendArgument = ...,
            n_workers: int | float | None = ..., shutdown_mode: ShutdownMode = ...) -> SingularMapping[T]: ...


@overload
def ultimap(func: Callable[..., T], inputs: Iterable, *, ordered: bool = ..., buffering: int | None = ...,
            batch_size: int = ..., errors: _ErrorNotReturn = ..., timeout: float | None = ...,
            return_key: Literal['idx'], recursive: bool = ..., backend: BackendArgument = ...,
            n_workers: int | float | None = ...,
            shutdown_mode: ShutdownMode = ...) -> SingularMapping[tuple[int, T]]: ...


@overload
def ultimap(func: Callable[..., T], inputs: Iterable[S], *, ordered: bool = ..., buffering: int | None = ...,
            batch_size: int = ..., errors: _ErrorNotReturn = ..., timeout: float | None = ...,
            return_key: Literal['input'], recursive: bool = ..., backend: BackendArgument = ...,
            n_workers: int | float | None = ...,
            shutdown_mode: ShutdownMode = ...) -> SingularMapping[tuple[S, T]]: ...


def ultimap(func: Callable, inputs: Iterable, *, ordered: bool = False, buffering: int | None = None,
            batch_size: int = 1, errors: Error = 'raise', timeout: float | None = None,
            return_key: ReturnKey = 'none', recursive: bool = False, backend: BackendArgument = "multiprocessing",
            n_workers: int | float | None = None, shutdown_mode: ShutdownMode = 'auto') -> SingularMapping:
    """
    A one-liner shortcut for creating a single-use Workforce and using it to map a function over several inputs.

    Thus, the call:
        results = list(ultimap(func, inputs))

    is roughly equivalent to:
        with Workforce() as wf:
            results = list(wf.map(func, inputs))

    Refer to `Workforce` and `Workforce.map` for full documentation.

    See Also
    --------
    Workforce
    Workforce.map
    """
    # TODO: can we also optimize n_workers here according to n_inputs?
    wf = Workforce(backend, n_workers, shutdown_mode)
    if recursive:
        func, inputs = make_recursive(func, inputs, wf.backend)
    return SingularMapping(wf, func, inputs, ordered, buffering, batch_size, errors, timeout, return_key)



class Args:
    """
    Canonized representation of task arguments.

    This class is used internally by the library, but can also be provided as a valid argument by users.
    It is specifically necessary when providing a tuple as the single argument for a task's function.

    Parameters
    ----------
    *args
        Positional arguments.
    **kwargs
        Keyword arguments.

    Attributes
    ----------
    raw
    """
    def __init__(self, *args, **kwargs):
        if '_created_from' in kwargs:
            self._created_from = kwargs.pop('_created_from')
        self.args = args
        self.kwargs = kwargs

    @property
    def raw(self):
        """Raw argument that was wrapped in this Args wrapper. Used internally."""
        return getattr(self, '_created_from', self)

    @classmethod
    def canonize(cls, item) -> 'Args':
        """
        Convert `item` into an `Args` object.

        The method provides convenient sugaring: If the given `item` is a tuple, it is automatically unpacked.
        In any case, the original `item` is saved and available for retrieval later on via the `raw` attribute.

        Parameters
        ----------
        item : Args, tuple or any other object
            - Args:      If the user has already provided an Args instance, returns the instance itself as is.
            - tuple:     In the special case `item` is a tuple, unpacks the tuple.
            - Otherwise: Used as a single input argument.

        Returns
        -------
        Args
            Canonized object representing the given `item` as arguments for a task.
        """
        if isinstance(item, cls):
            assert item.raw is item
            return item
        elif isinstance(item, tuple):
            return cls(*item, _created_from=item)
        else:
            return cls(item, _created_from=item)



"""
Helpers and utilities.
"""
import logging
import itertools
import threading
from typing import TypeVar
from collections.abc import Iterable


T = TypeVar('T')


def batches(iterable: Iterable[T], size: int) -> Iterable[Iterable[T]]:
    """
    Iterate over the given iterable in batches of fixed size.
    The last batch may be smaller.
    """
    it = iter(iterable)
    while True:
        batchiter = itertools.islice(it, size)
        try:
            next_item = next(batchiter)
        except StopIteration:
            return
        yield itertools.chain([next_item], batchiter)


class class_logger:  # noqa
    def __get__(self, obj, cls):
        return logging.getLogger(cls.__module__).getChild(cls.__name__)


class SyncCounter:
    """
    Thread-safe counter.

    A class holding a numeric value and capable of performing atomic post-increment on it.

    Parameters
    ----------
    next_value : int, default 0
        The first value to be returned from `get_next_value`.
    """
    def __init__(self, next_value: int = 0):
        self.next_value = next_value
        self._lock = threading.Lock()

    def get_next_value(self) -> int:
        """
        Get the next value of the counter.

        Safely increment the value by 1 and return the value before incrementation.

        Returns
        -------
        int
            The value of the counter.
        """
        with self._lock:
            result = self.next_value
            self.next_value += 1
        return result



"""
Classes for communicating between the Workforce and the workers.
"""
import os
import sys
import signal
import threading
import traceback
from typing import TYPE_CHECKING
from collections.abc import Iterable, Mapping, Callable

from .args import Args
from ._registry import RegistryKey
if TYPE_CHECKING:
    from .workforce import Workforce


# TODO: convert these into module functions?
class WorkerAPI:
    """
    A collection of methods for interacting with the workers.

    Attributes
    ----------
    tl : threading.local
        Data local to each thread.
        Specifically, `tl.records` is a dictionary storing the functions accessible to the workers.
    """
    tl = threading.local()

    @classmethod
    def initializer(cls, workforce_id: int, functions: Mapping[RegistryKey, Callable], manager_pid: int) -> None:
        """
        Initialize the worker.

        This is called by each worker once, when it is spun up, before any tasks are performed.

        Parameters
        ----------
        workforce_id : int
            The ID of the Workforce of which this worker is a part.
        functions : mapping
            Maps keys to callable objects. These are the functions accessible to the worker in order to perform tasks.
        manager_pid : int
            The ID of the process that spun the worker up.

        Returns
        -------
        None
        """
        if not hasattr(cls.tl, "records"):
            cls.tl.records = {}
        elif workforce_id in cls.tl.records:
            raise RuntimeError(f"WorkerAPI re-initialized for the same Workforce id {workforce_id}")
        cls.tl.records[workforce_id] = functions
        if manager_pid != os.getpid():
            # we're a spawn, protect ourselves from SIGINT
            signal.signal(signal.SIGINT, signal.SIG_IGN)

    @staticmethod
    def init_args(workforce: "Workforce") -> tuple[int, Mapping[RegistryKey, Callable], int]:
        """
        Get the required arguments for initialization of the worker.
        This is a convenience function.

        Parameters
        ----------
        workforce : Workforce
            The Workforce initializing the worker.

        Returns
        -------
        tuple
            Arguments for the `initializer` method.
        """
        return workforce.workforce_id, workforce.func_registry.get_deserializer(), os.getpid()

    @classmethod
    def _get_func(cls, workforce_id: int, key: RegistryKey) -> Callable:
        return cls.tl.records[workforce_id][key]

    @classmethod
    def execute_one(cls, workforce_id: int, key: RegistryKey, args: Args):
        """
        Execute a single task.

        Parameters
        ----------
        workforce_id : int
            The ID of the Workforce submitting the task.
        key : str or int
            Key identifying the function for executing the task.
        args : Args
            Arguments for the function.

        Returns
        -------
        Result of executing the function with given `args`.
        """
        assert isinstance(args, Args)
        func = cls._get_func(workforce_id, key)
        return func(*args.args, **args.kwargs)

    @classmethod
    def execute(cls, workforce_id: int, key: RegistryKey, args_iter: Iterable[Args]) -> list:
        """
        Execute several tasks as a batch.
        The tasks must share the same function for execution.

        Parameters
        ----------
        workforce_id : int
            The ID of the Workforce submitting the tasks.
        key : str or int
            Key identifying the function for executing all the tasks.
        args_iter : iterable of Args
            Each item is the arguments for a single invocation of the function.

        Returns
        -------
        list
            Results of all function invocations.
            If an exception was raised by the function, the result is replaced by an `ErrorResult` object.
        """
        func = cls._get_func(workforce_id, key)
        results = []
        for args in args_iter:
            assert isinstance(args, Args)
            try:
                result = func(*args.args, **args.kwargs)
            except BaseException:  # noqa
                result = ErrorResult.from_current_exc()
            results.append(result)
        return results


class ErrorResult:
    """
    Holds an exception and its traceback, allowing safe pickling while retaining the traceback text.
    """
    def __init__(self, exc_info):
        self.exc = exc_info[1]
        self.exc.__traceback__ = exc_info[2]

    def __getstate__(self):
        tb_text = f"\n'''\n{''.join(traceback.format_exception(type(self.exc), self.exc, self.exc.__traceback__))}'''"
        return self.exc, tb_text

    def __setstate__(self, state):
        exc, tb_text = state
        self.exc = exc
        self.exc.__cause__ = _WorkerTraceback(tb_text)

    @classmethod
    def from_current_exc(cls) -> "ErrorResult":
        return cls(sys.exc_info())


class _WorkerTraceback(Exception):
    def __init__(self, tb_text: str):
        self.tb_text = tb_text

    def __str__(self):
        return self.tb_text



"""
This module contains types that are outward-facing, to be imported and used by users of the package.
"""
from typing import Literal, TypeAlias

from .backend import BackendArgument


ReturnKey: TypeAlias = Literal['none', 'idx', 'input']
Error: TypeAlias = Literal['raise', 'ignore', 'log', 'return']



import os
import time
import queue
import threading
import multiprocessing as mp
from typing import ClassVar, TypeVar, Generic
from collections.abc import Iterable, Callable
from multiprocessing.managers import SyncManager

from .args import Args
from .backend import Backend, MultiprocessingBackend, ThreadingBackend, InlineBackend


T = TypeVar("T")
S = TypeVar("S")


def make_recursive(
        func: Callable[..., T],
        inputs: Iterable[S],
        backend: Backend) -> tuple[Callable[..., T], Iterable[S]]:
    recursion = Recursion(func, inputs, backend)
    return recursion.func, recursion.inputs


class Recursion(Generic[T, S]):
    _MP_MANAGER: ClassVar[SyncManager | None] = None

    def __init__(self, func: Callable[..., T], inputs: Iterable[S], backend: Backend,
                 *, interval: float = 0.1):
        self._orig_func = func
        self._orig_inputs = inputs
        self._interval = interval
        self._queue = self._make_queue(backend)
        self._safe_data = _SafeData(backend)

    @classmethod
    def _make_queue(cls, backend: Backend):
        if isinstance(backend, MultiprocessingBackend):
            if cls._MP_MANAGER is None:
                cls._MP_MANAGER = mp.Manager()
            return cls._MP_MANAGER.Queue()
        elif isinstance(backend, (ThreadingBackend, InlineBackend)):
            return queue.Queue()
        else:
            raise ValueError(f"backend {backend} is not supported")

    # interface for func callback to add tasks, and for its wrapper to mark tasks as done
    # these two should be called within the worker

    def add_input(self, *args, **kwargs) -> None:
        # we keep the semantics of ultima's input specification
        if kwargs or len(args) != 1:
            task_args = Args(*args, **kwargs)
        else:
            # this supports all canonize-able forms of input: a single argument, a tuple to expand, and an Args instance
            task_args = args[0]
        self._queue.put(task_args)
        self._safe_data.queue_size_inc()

    def _mark_task_done(self):
        self._safe_data.tasks_done_inc()

    # adapters for ultima's interface of inputs, func and backend (stored as member variable)

    @property
    def inputs(self) -> Iterable[S]:
        self._safe_data.set_owner()
        source_iterator = iter(self._orig_inputs)
        source_depleted = False
        while True:
            if self._safe_data.outstanding_tasks == 0 and self._safe_data.queue_size == 0 and source_depleted:
                return
            if self._safe_data.queue_size > 0:
                self._safe_data.queue_size_dec()
                self._safe_data.tasks_scheduled_inc()
                yield self._queue.get()
                continue
            if not source_depleted:
                try:
                    next_input = next(source_iterator)
                except StopIteration:
                    source_depleted = True
                else:
                    self._queue.put(next_input)
                    self._safe_data.queue_size_inc()
                continue
            # we have no inputs to yield, but our outstanding tasks might still add new tasks, so we sleep and retry
            time.sleep(self._interval)
            continue

    def func(self, *args, **kwargs) -> T:
        try:
            return self._orig_func(*args, **kwargs, add_input=self.add_input)
        finally:
            self._mark_task_done()


class _SafeData:
    # warning: the caller is responsible to keep backend alive for the span of this object's lifecycle.
    # when a backend gets GCed, its manager gets shut down and the data invalidates.
    def __init__(self, backend: Backend):
        # we avoid a lock by giving each process/thread combination its own counters
        self._multi_queue_size: dict[tuple[int, int], int] = backend.dict()
        self._multi_tasks_done: dict[tuple[int, int], int] = backend.dict()
        self._tasks_scheduled = 0
        self._owner_key = None

    def queue_size_inc(self):
        key = self._caller_key()
        self._multi_queue_size[key] = self._multi_queue_size.get(key, 0) + 1

    def queue_size_dec(self):
        self._check_owner()
        self._multi_queue_size[self._owner_key] = self._multi_queue_size.get(self._owner_key, 0) - 1

    def tasks_scheduled_inc(self):
        self._check_owner()
        self._tasks_scheduled += 1

    def tasks_done_inc(self):
        key = self._caller_key()
        self._multi_tasks_done[key] = self._multi_tasks_done.get(key, 0) + 1

    @property
    def queue_size(self):
        self._check_owner()
        return sum(self._multi_queue_size.values())

    @property
    def outstanding_tasks(self):
        return self._tasks_scheduled - sum(self._multi_tasks_done.values())

    def set_owner(self):
        if self._owner_key is not None:
            raise Exception("owner has already been set")
        self._owner_key = self._caller_key()

    def _check_owner(self):
        if self._owner_key is None:
            raise Exception("the owner has not been set")
        if not self._caller_key() == self._owner_key:
            raise Exception("only the owner of this object can call this method")

    @staticmethod
    def _caller_key():
        return os.getpid(), threading.get_native_id()



import logging

from .workforce import Workforce, ultimap
from .args import Args
from .backend import (
    InlineBackend,
    ThreadingBackend,
    MultiprocessingBackend,
    MultiprocessingForkBackend,
    MultiprocessingSpawnBackend,
)

logging.getLogger(__name__).addHandler(logging.NullHandler())



"""
Various execution backends, all using a single interface.
"""
import os
import sys
import math
import signal
import threading
import multiprocessing
import multiprocessing.managers
import concurrent.futures
from functools import partial
from abc import ABC, abstractmethod
from typing import TypeVar, Literal, overload, TypeAlias

import dill

from .inline import InlineExecutor


#
# Interface
#

class Backend(ABC):
    """
    Abstract base class for an execution backend, providing a common interface.

    Attributes
    ----------
    executor_shutdown_nowait_allowed
    dict
    Executor
    NAME : str
        The name of the backend. Can be used in `get_backend` to get an instance of the backend.
    """
    def __del__(self):
        self.shutdown()

    @property
    def executor_shutdown_nowait_allowed(self) -> bool:
        """Indicates whether we should call executor.shutdown with wait=False."""
        return True

    @classmethod
    @abstractmethod
    def parse_n_workers(cls, n_workers) -> int:
        """
        Parse the `n_workers` parameters in the context of the relevant backend
        and return the actual number of workers to use.

        Parameters
        ----------
        n_workers : int, float or None
            Generalized number of workers. Allowed types depend on the backend.

        Returns
        -------
        int
            Actual number of workers to use.
        """
        raise NotImplementedError()

    @staticmethod
    def serialize(item):
        """
        Serialize an item in a way that's appropriate for the backend.

        Parameters
        ----------
        item
            Item to be serialized.

        Returns
        -------
        object
            Serialized `item`.
        """
        return item

    @staticmethod
    def deserialize(item):
        """
        Inverse of the `serialize` method.

        Parameters
        ----------
        item
            An object previously returned by `serialize`.

        Returns
        -------
        object
            Original object.
        """
        return item

    @property
    def dict(self):
        """A mapping class appropriate for use with this backend."""
        return dict

    @property
    @abstractmethod
    def Executor(self):
        """An Executor class appropriate for use with this backend."""
        raise NotImplementedError()

    def shutdown(self):
        """Shut down any resources this backend uses."""
        pass


BackendType = TypeVar('BackendType', bound=Backend)
BackendArgument: TypeAlias = str | Backend | type[Backend]


@overload
def get_backend(name_or_backend: BackendType | type[BackendType]) -> BackendType: ...
@overload
def get_backend(name_or_backend: str) -> Backend: ...


def get_backend(name_or_backend: BackendArgument):
    """
    Get a Backend object by name or, if already providing a backend class or object,
    returns an instance of the backend.

    Parameters
    ----------
    name_or_backend : str, Backend or Backend class
        Identifies the required backend.

    Returns
    -------
    Backend
        If `name_or_backend` is already a Backend instance, returns the same object.
        Otherwise, returns a new instance of the required backend.
    """
    if isinstance(name_or_backend, Backend):
        return name_or_backend
    elif isinstance(name_or_backend, type) and issubclass(name_or_backend, Backend):
        backend_class = name_or_backend
    elif isinstance(name_or_backend, str):
        backend_classes = [i for i in globals().values() if isinstance(i, type) and issubclass(i, Backend)]
        for backend_class in backend_classes:
            if getattr(backend_class, "NAME", None) == name_or_backend:
                break
        else:
            names = [name for cls in backend_classes if (name := getattr(cls, "NAME", None)) is not None]
            raise ValueError(f"invalid backend name '{name_or_backend}', should be one of {names}")
    else:
        raise TypeError(f"invalid backend requested: {name_or_backend}")
    return backend_class()


#
# Backend implementations
#

class InlineBackend(Backend):
    """
    A backend for executing tasks inline, i.e. without any execution pool.
    """
    NAME = "inline"

    @classmethod
    def parse_n_workers(cls, n_workers: Literal[0, None]) -> Literal[0]:
        if n_workers not in (0, None):
            raise ValueError(f"{cls.__name__} does not support {n_workers=}, only zero")
        return 0

    @property
    def Executor(self):
        return InlineExecutor


class ThreadingBackend(Backend):
    """
    A backend for executing tasks using multithreading.
    """
    NAME = "threading"

    @classmethod
    def parse_n_workers(cls, n_workers: int) -> int:
        if n_workers is None:
            raise ValueError(f"{cls.__name__} does not provide a default number of workers")
        if n_workers <= 0 or type(n_workers) is float:
            raise ValueError(f"{cls.__name__} does not support non-positive or floating-point worker numbers")
        return n_workers

    @property
    def Executor(self):
        return concurrent.futures.ThreadPoolExecutor


class MultiprocessingBackend(Backend):
    """
    A backend for executing tasks using multiprocessing.

    Attributes
    ----------
    manager
    CONTEXT : str or None
        Name of the multiprocessing context used by this backend.
    """
    NAME = "multiprocessing"
    CONTEXT: str | None = None

    def __init__(self):
        self._manager = None
        self._lock = threading.Lock()

    @property
    def _ctx(self):
        if self.CONTEXT is None:
            return None
        else:
            return multiprocessing.get_context(self.CONTEXT)

    @property
    def manager(self) -> multiprocessing.managers.SyncManager:
        if self._manager is None:
            with self._lock:
                if self._manager is None:
                    manager = multiprocessing.managers.SyncManager(ctx=self._ctx)
                    manager.start(self._mgr_init)
                    self._manager = manager
        return self._manager

    @staticmethod
    def _mgr_init():
        signal.signal(signal.SIGINT, signal.SIG_IGN)

    @property
    def executor_shutdown_nowait_allowed(self):
        # there's a multiprocessing bug that was fixed in py 3.9 when shutting down with wait=False
        # https://github.com/python/cpython/commit/a5cbab552d294d99fde864306632d7e511a75d3c
        return sys.version_info >= (3, 9)

    @classmethod
    def parse_n_workers(cls, n_workers: int | float | None) -> int:
        """
        Parse the `n_workers` parameters and return the actual number of workers to use.

        Parameters
        ----------
        n_workers : int, float or None
            Generalized number of workers. Different values have different meanings:
            - None:         Use total number of CPUs.
            - Positive int: Actual number of workers.
            - Negative int: Use n_CPUs + `n_workers` + 1, i.e. -1 means n_CPUs, -2 means 1 fewer, etc.
            - float:        Fraction of the number of CPUs. Must be positive.

        Returns
        -------
        int
            Actual number of workers to use.

        Raises
        ------
        ValueError
            If `n_workers` equals zero (either int or float).
        TypeError
            If `n_workers` is of an unexpected type.
        NotImplementedError
            If the number of CPUs is required but cannot be computed.
        """
        if n_workers is None:
            n_workers = -1
        if isinstance(n_workers, int):
            if n_workers < 0:
                cpu_count = os.cpu_count()
                if cpu_count is None:
                    raise NotImplementedError("cannot determine number of CPUs")
                n_workers = max(cpu_count + 1 + n_workers, 1)
            elif n_workers == 0:
                raise ValueError(f"{cls.__name__} does not support {n_workers=}")
        elif isinstance(n_workers, float):
            if n_workers <= 0:
                raise ValueError(f"{cls.__name__} does not support non-positive float {n_workers=}")
            n_workers = math.ceil(multiprocessing.cpu_count() * n_workers)
        else:
            raise TypeError(f"{cls.__name__} does not support {n_workers=}")
        return n_workers

    @staticmethod
    def serialize(item):
        return dill.dumps(item)

    @staticmethod
    def deserialize(item):
        return dill.loads(item)

    @property
    def dict(self):
        return self.manager.dict

    @property
    def Executor(self):
        executor = concurrent.futures.ProcessPoolExecutor
        if (ctx := self._ctx) is not None:
            executor = partial(executor, mp_context=ctx)
        return executor

    def shutdown(self):
        if (manager := self._manager) is not None:
            self._manager = None
            manager.shutdown()


class MultiprocessingSpawnBackend(MultiprocessingBackend):
    NAME = 'multiprocessing-spawn'
    CONTEXT = 'spawn'


class MultiprocessingForkBackend(MultiprocessingBackend):
    NAME = 'multiprocessing-fork'
    CONTEXT = 'fork'


class MultiprocessingForkserverBackend(MultiprocessingBackend):
    NAME = 'multiprocessing-forkserver'
    CONTEXT = 'forkserver'



"""
Mock implementations of Executor and Future for inline execution,
i.e. without any execution pool or parallelization.
"""
from concurrent.futures import Executor, Future


class InlineExecutor(Executor):
    """
    A mock executor, executing using `InlineFuture` objects.
    """
    def __init__(self, max_workers=None, initializer=None, initargs=()):
        if initializer is not None:
            initializer(*initargs)

    def submit(self, fn, *args, **kwargs):
        return InlineFuture(fn, args, kwargs)


class InlineFuture(Future):
    """
    A mock Future, executing inline upon request of result.
    """
    MOCK_RESULT = object()

    def __init__(self, fn, args, kwargs):
        self._result = None
        self.func = fn
        self.args = args
        self.kwargs = kwargs
        super().__init__()
        self.set_result(self.MOCK_RESULT)

    def result(self, timeout=None):
        if self._result is self.MOCK_RESULT:
            self._result = self.func(*self.args, **self.kwargs)
        return super().result(timeout=timeout)



import time
import threading
import concurrent.futures
from typing import Generic, TypeVar, Any
from collections.abc import Iterable, Iterator

from .utils import class_logger


T = TypeVar("T")


class KeyedFuture(Generic[T]):
    def __init__(self, key: T, future: concurrent.futures.Future):
        self.key = key
        self.future = future


class BufferedFutureResolver(Generic[T]):
    """
    A generator of results from futures, supporting buffering and timeout.

    This class consumes futures in a background thread and provides iteration on the resolved results.
    Consumption of input only starts when starting iteration on the results.
    Errors are handled according to the user's preference. If aborting prematurely, unresolved futures are canceled.

    Parameters
    ----------
    keyed_futures : iterable of KeyedFuture
        The futures to resolve, each paired with a key via the KeyedFuture helper class.
    abort_on_error : bool
        Controls the behavior when an exception is raised when obtaining a result:
        - True:  Abort the iteration and raise the exception.
        - False: Return the exception object instead of the result.
    ordered : bool
        Whether the order of the outputs should correlate to the order of the inputs in `keyed_futures`.
    buffering : int, optional
        Limits the number of futures extracted from `keyed_futures`. When the buffering limit is reached, wait
        until some futures are resolved (finished) before extracting more.
        By default, no limit is set.
    timeout : float, optional
        A global timeout, in seconds, on the entire procedure of resolving all the futures.
        The clock starts as soon as iteration over the results starts.
    name : str, optional
        Name for the feeder thread.
    """
    SYNC_TIMEOUT = 0.2

    logger = class_logger()

    def __init__(self, keyed_futures: Iterable[KeyedFuture[T]], abort_on_error: bool, ordered: bool,
                 buffering: int | None = None, timeout: float | None = None, name: str | None = None):
        self.logger.debug("created")
        self.keyed_futures = iter(keyed_futures)
        self.abort_on_error = abort_on_error
        self.ordered = ordered
        self.capacity = buffering
        self.timeout = timeout
        self.name = name
        self.active = True
        self._start_time = None
        self._feeder_error = None

        self.lock = threading.Lock()
        self.evt_new_future = threading.Event()
        self.evt_new_capacity = threading.Event()
        self.futures: dict[concurrent.futures.Future, T] = {}

        self.feeder_thread = threading.Thread(
            target=self._feeder_thread_impl,
            daemon=True,
            name=None if self.name is None else f"{self.name}-FeederThread"
        )

    def __iter__(self) -> Iterator[tuple[T, Any]]:
        self._start_time = time.monotonic()
        self.feeder_thread.start()
        yield from self._futures_iter()

    # dev note: this code runs in a single-use thread
    # WARNING: this is a daemon thread, do not assume resource cleanup or try-finally mechanisms work
    def _feeder_thread_impl(self):
        self.logger.debug("feeder thread running")
        # once a future has been created, we're responsible for it until we've put it into the future set.
        # TODO: Replace "while True" with "while self.active"
        while True:
            try:
                keyed_future = next(self.keyed_futures)
            except StopIteration:
                break
            except concurrent.futures.BrokenExecutor as exc:
                self.logger.warning("feeder thread aborting due to executor failure in future generation")
                self._feeder_error = exc
                return
            except BaseException as exc:
                if isinstance(exc, Exception):
                    err = "error"
                    log_method = self.logger.info
                else:
                    err = "base exception"
                    log_method = self.logger.exception
                log_method(f"{err} thrown from future generator, marking broken and exiting thread")
                self._feeder_error = exc
                return
            try:
                # we'll keep trying to put it into the future set until successful or until we're asked to stop feeding
                while True:
                    if not self.active:
                        # by definition, we always hold an untracked future at this point, so it's on us to cancel it
                        self.logger.debug("feeder thread exiting since BFR is not active (canceling future)")
                        was_canceled = keyed_future.future.cancel()
                        if was_canceled:
                            # work around gh-94440 by sleeping a tiny bit before potentially shutting down the executor
                            # solves issue #3
                            time.sleep(0.05)
                        return
                    if self._has_capacity():
                        with self.lock:
                            self.futures[keyed_future.future] = keyed_future.key
                        self.evt_new_future.set()
                        break
                    if self.evt_new_capacity.wait(timeout=self.SYNC_TIMEOUT):
                        self.evt_new_capacity.clear()
            except BaseException as exc:
                self.logger.exception("internal error in feeder thread, marking broken and raising to die")
                self._feeder_error = exc
                raise
        self.logger.debug("feeder thread done")

    def _has_capacity(self) -> bool:
        return self.capacity is None or len(self.futures) < self.capacity

    # dev note: this core generator happens in the main thread
    def _futures_iter(self) -> Iterable[tuple[T, Any]]:
        try:
            while (n_futures := len(self.futures)) or self.feeder_thread.is_alive() and self._feeder_error is None:
                self._check_timeout()
                if not n_futures:
                    if not self.evt_new_future.wait(timeout=self.SYNC_TIMEOUT):
                        continue
                    self.evt_new_future.clear()
                    if not self.futures:
                        continue
                self._check_timeout()
                with self.lock:
                    if self.ordered:
                        futures = [next(iter(self.futures))]
                    else:
                        futures = list(self.futures)
                time_remaining = self._check_timeout()
                done, not_done = concurrent.futures.wait(futures, time_remaining, concurrent.futures.FIRST_COMPLETED)
                for done_future in done:
                    self._check_timeout()
                    with self.lock:
                        key = self.futures.pop(done_future)
                    self.evt_new_capacity.set()
                    self._check_timeout()
                    try:
                        yield key, done_future.result()
                    except concurrent.futures.TimeoutError as exc:
                        # replace with a standard TimeoutError
                        raise TimeoutError("future timed out") from exc
                    except Exception as exc:
                        if self.abort_on_error:
                            raise
                        yield key, exc
        except GeneratorExit:
            # we were asked nicely to terminate, no need to report an error
            raise
        except:
            self.logger.exception("error raised in _futures_iter")
            raise
        finally:
            self.terminate()
        # noinspection PyUnreachableCode
        if self._feeder_error is not None:
            raise self._feeder_error

    def _check_timeout(self) -> float | None:
        if self.timeout is None:
            return None
        elapsed = time.monotonic() - self._start_time
        remaining = self.timeout - elapsed
        if remaining <= 0:
            msg = f"Timed out after {elapsed:g} seconds (timeout={self.timeout})"
            self.logger.warning(msg)
            raise TimeoutError(msg)
        return remaining

    def terminate(self) -> None:
        if not self.active:
            return
        self.logger.debug("cleaning up futures")
        self.active = False
        # we could be terminated due to an error in the feeder thread, in which case no need to join
        if self.feeder_thread.is_alive() and self.feeder_thread is not threading.current_thread():
            self.feeder_thread.join()
        if n := len(self.futures):
            self.logger.debug(f"canceling {n} pending futures")
        any_future_canceled = False
        for future in self.futures.keys():
            any_future_canceled |= future.cancel()
        if any_future_canceled:
            # work around gh-94440 by sleeping a tiny bit before proceeding to (probably) shut down the executor
            # solves issue #2
            time.sleep(0.05)
        self.logger.debug("cleaned up futures")



import hashlib
from typing import TypeVar, Generic, TypeAlias
from collections.abc import Callable, Iterator, Mapping

from .backend import Backend


T = TypeVar("T")
KT = TypeVar("KT")
VT_co = TypeVar('VT_co', covariant=True)
RegistryKey: TypeAlias = str | int


class DeserializerMapping(Mapping[KT, VT_co]):
    """
    A wrapper for a mapping, which also deserializes the values upon access.

    In addition, values are cached when accessed.
    """
    def __init__(self, items: Mapping[KT, T], deserializer: Callable[[T], VT_co]):
        self._items = items
        self._deserializer = deserializer
        self._cache: dict[KT, VT_co] = {}

    def __getitem__(self, key: KT) -> VT_co:
        if key not in self._cache:
            self._cache[key] = self._deserializer(self._items[key])
        return self._cache[key]

    def __len__(self) -> int:
        return len(self._items)

    def __iter__(self) -> Iterator[KT]:
        return iter(self._items)


class SerializedItemsRegistry(Generic[T]):
    """
    A registry of serialized items, accessible from both the Workforce side and the workers.

    The registry is implemented using a Mapping appropriate for the backend used, and the serialization is also
    performed according to the backend.

    Parameters
    ----------
    backend : Backend
        The backend to use for communication.
    """
    def __init__(self, backend: Backend):
        self.backend = backend
        self.items = backend.dict()

    def register(self, item: T) -> RegistryKey:
        s_item = self.backend.serialize(item)
        key = self._make_key(s_item)
        if key not in self.items:
            self.items[key] = s_item
        return key

    @staticmethod
    def _make_key(serialized_item) -> RegistryKey:
        if isinstance(serialized_item, bytes):
            return hashlib.sha1(serialized_item).hexdigest()
        else:
            return id(serialized_item)

    def get_deserializer(self) -> DeserializerMapping[RegistryKey, T]:
        return DeserializerMapping(self.items, self.backend.deserialize)
